# type: ignore

# Test schema generation
import json
import os

import pytest

import event_model
from event_model.documents import ALL_DOCUMENTS
from event_model.documents.generate.typeddict_to_schema import typeddict_to_schema

SCHEMA_PATH = event_model.__path__[0] + "/schemas/"


@pytest.mark.parametrize("typed_dict_class", ALL_DOCUMENTS)
def test_generated_json_matches_typed_dict(typed_dict_class, tmpdir):
    typeddict_to_schema(typed_dict_class, schema_dir=tmpdir)
    file_name = os.listdir(tmpdir)[0]
    generated_file_path = os.path.join(tmpdir, file_name)
    old_file_path = os.path.join(SCHEMA_PATH, file_name)

    with open(generated_file_path) as generated_file, open(old_file_path) as old_file:
        try:
            assert json.load(generated_file) == json.load(old_file)
        except AssertionError:
            raise Exception(
                f"`{typed_dict_class.__name__}` can generate a json schema, but "
                f"it doesn't match the schema in `{SCHEMA_PATH}`. Did you forget "
                "to run `python event_model/documents/generate` after changes "
                f"to `{typed_dict_class.__name__}`?"
            )



"""
A wrapper used to patch out schema generation utilities.
"""

from typing import TYPE_CHECKING, Any, Dict, List, Union

pydantic_version = None


def Field(*args, **kwargs): ...


class BaseModel: ...


class TypeAdapter:
    def __init__(self, *args, **kwargs):
        self.by_alias = True
        ...

    def json_schema(self, *args, **kwargs): ...


class GenerateJsonSchema: ...


def to_snake(*args, **kwargs): ...


DataFrameForDatumPage = List[str]

DataFrameForEventPage = Dict[str, List]

DataFrameForFilled = Dict[str, List[Union[bool, str]]]


class DataType: ...


def add_extra_schema(*args, **kwargs):
    def inner(cls):
        return cls

    return inner


# Dictionary for patching in schema post generation
extra_schema = {}  # type: ignore

if not TYPE_CHECKING:
    try:
        import pydantic

        if pydantic.__version__ < "2":
            raise ImportError

        pydantic_version = pydantic.__version__

        # Root models for root definitions:
        # we want some types to reference definitions in the
        # schema
        BaseModel = pydantic.BaseModel
        ConfigDict = pydantic.ConfigDict
        Field = pydantic.Field
        RootModel = pydantic.RootModel
        TypeAdapter = pydantic.TypeAdapter
        GenerateJsonSchema = pydantic.json_schema.GenerateJsonSchema
        from pydantic.alias_generators import to_snake  # noqa

        class Config(ConfigDict): ...

        class DataFrameForDatumPage(RootModel):
            root: List[str] = Field(alias="Dataframe")

        class DataFrameForEventPage(RootModel):
            root: Dict[str, List] = Field(alias="Dataframe")

        class DataFrameForFilled(RootModel):
            root: Dict[str, List[Union[bool, str]]] = Field(alias="DataframeForFilled")

        class DataType(RootModel):
            root: Any = Field(alias="DataType")

        def add_extra_schema(schema: Dict) -> Dict:
            def inner(cls):
                extra_schema[cls] = schema
                return cls

            return inner

    # If pydantic is not installed (e.g the install isn't [dev]),
    # or pydantic v1 is being used, then we expect to be able to
    # run event-model, just not the schema generation code.
    except (ModuleNotFoundError, ImportError):
        # None of the dummy functions/classes should have been overwritten.
        assert pydantic_version is None



from typing import Tuple, Type, Union

# flake8: noqa
from event_model.documents.datum import Datum
from event_model.documents.datum_page import DatumPage
from event_model.documents.event import Event
from event_model.documents.event_descriptor import EventDescriptor
from event_model.documents.event_page import EventPage
from event_model.documents.resource import Resource
from event_model.documents.run_start import RunStart
from event_model.documents.run_stop import RunStop
from event_model.documents.stream_datum import StreamDatum
from event_model.documents.stream_resource import StreamResource

DocumentType = Union[
    Type[Datum],
    Type[DatumPage],
    Type[Event],
    Type[EventDescriptor],
    Type[EventPage],
    Type[Resource],
    Type[RunStart],
    Type[RunStop],
    Type[StreamDatum],
    Type[StreamResource],
]

ALL_DOCUMENTS: Tuple[DocumentType, ...] = (
    Datum,
    DatumPage,
    Event,
    EventDescriptor,
    EventPage,
    Resource,
    RunStart,
    RunStop,
    StreamDatum,
    StreamResource,
)


__all__ = [
    "Datum",
    "DatumPage",
    "Event",
    "EventDescriptor",
    "EventPage",
    "Resource",
    "RunStart",
    "RunStop",
    "StreamDatum",
    "StreamResource",
    "DocumentType",
]



import json
from collections import OrderedDict
from pathlib import Path
from typing import Dict, Optional

from event_model.documents import ALL_DOCUMENTS, DocumentType
from event_model.documents.generate.type_wrapper import (
    GenerateJsonSchema,
    TypeAdapter,
    extra_schema,
    to_snake,
)


def sort_alphabetically(schema: Dict) -> Dict:
    """Sorts the schema alphabetically by key name, exchanging the
    properties dicts for OrderedDicts"""
    schema = OrderedDict(sorted(list(schema.items()), key=lambda x: x[0]))

    return schema


SortOrder = {
    "title": 0,
    "description": 1,
    "type": 2,
    "$defs": 3,
    "properties": 4,
    "required": 5,
    "additionalProperties": 6,
    "patternProperties": 7,
}


def sort_schema(document_schema: Dict) -> Dict:
    assert isinstance(document_schema, dict)
    document_schema = OrderedDict(
        sorted(
            list(document_schema.items()),
            key=lambda x: SortOrder.get(x[0], len(SortOrder)),
        )
    )

    for key in document_schema:
        if key in ("$defs", "properties", "required"):
            if isinstance(document_schema[key], dict):
                document_schema[key] = sort_alphabetically(document_schema[key])
                for key2 in document_schema[key]:
                    if isinstance(document_schema[key][key2], dict):
                        document_schema[key][key2] = sort_schema(
                            document_schema[key][key2]
                        )
            elif isinstance(document_schema[key], list):
                document_schema[key].sort()

    return document_schema


def add_extra_schema(document_schema: Dict, extra_schema: Dict):
    """
    Used to add user written schema to autogenerated schema.

    Once https://github.com/pydantic/pydantic/issues/8698 is fixed we can
    use the following instead of this function:

    class document_type(document_type):
        __pydantic_config__ = ConfigDict(
            title=document_type.__name__,
            json_schema_extra=extra_schema.get(document_type, None)
        )
    """
    for key in extra_schema:
        if key not in document_schema:
            document_schema[key] = extra_schema[key]

        elif not isinstance(document_schema[key], type(extra_schema[key])):
            raise ValueError(
                f"Cannot merge {repr(document_schema[key])}:"
                f"{type(document_schema[key])} with {extra_schema[key]}:"
                f"{type(extra_schema[key])} in {document_schema['title']}[{key}]"
            )

        elif isinstance(document_schema[key], dict):
            add_extra_schema(document_schema[key], extra_schema[key])

        elif isinstance(document_schema[key], list):
            document_schema[key] += extra_schema[key]
        else:
            raise ValueError(
                f"{repr(document_schema[key])}:{type(document_schema[key])} in "
                f"{document_schema['title']}[{key}] is of unsupport type"
            )


def dump_json(dictionary: Dict, file_path: Path, mode="w"):
    with open(file_path, mode) as f:
        json.dump(dictionary, f, indent=4)


class _GenerateJsonSchema(GenerateJsonSchema):
    def generate(self, schema, mode="validation"):
        json_schema = super().generate(schema, mode=mode)
        json_schema["title"] = to_snake(json_schema["title"])
        json_schema["description"] = (
            json_schema["description"].replace("\n", " ").strip()
        )
        return json_schema


def typeddict_to_schema(
    document_type: DocumentType, schema_dir: Optional[Path] = None, sort: bool = True
) -> Dict:
    assert document_type in ALL_DOCUMENTS

    type_adapter = TypeAdapter(document_type)
    type_adapter.by_alias = True
    document_schema = type_adapter.json_schema(schema_generator=_GenerateJsonSchema)

    if sort:
        document_schema = sort_schema(document_schema)

    # Add the manually defined extra schema
    add_extra_schema(document_schema, extra_schema.get(document_type, {}))

    if schema_dir:
        file_path = schema_dir / f"{to_snake(document_schema['title'])}.json"

        # Check if the file has been updated
        if not file_path.exists():
            print(f"{str(file_path)} does not exist yet, writing")
            dump_json(document_schema, file_path)
        else:
            with open(file_path) as json_file:
                skip_writing = json.load(json_file) == document_schema

            if skip_writing:
                print(f"{document_schema['title']} is unchanged")
            else:
                print(
                    f"{document_schema['title']} has been changed, writing new schema"
                )
                dump_json(document_schema, file_path)

    return document_schema



from pathlib import Path

from event_model.documents import ALL_DOCUMENTS
from event_model.documents.generate.typeddict_to_schema import typeddict_to_schema

if __name__ == "__main__":
    schema_dir = Path(__file__).parent.parent.parent / "schemas"
    for document in ALL_DOCUMENTS:
        typeddict_to_schema(document, schema_dir)
